// -*- mode:C++; tab-width: 4 -*-
/*
    Multi-threaded C version fo email-fmt.py.
    Copyright (C) 2005 Grzegorz Adam Hankiewicz.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License,
    or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
    USA.
*/
#include "email-fmt.h"

#include "misc.h"
#include "shared-list.h"

#include <assert.h>
#include <ctype.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int _count_indentation(const char *line, const int soft);
static void _fork_line_block(char **lines, const int indentation);

static const char *_whitespace = " \f\n\r\t\v";
static const char *_indentation_characters = " \t>:;#%-";
static int _target_block_width = 70;

/* Internal functions. */
static void *_producer_thread(void *data);
static void *_consumer_thread(void *data);

/**[txh]********************************************************************

  Description: Real horsework of the program. Receives as parameter
  a boolean flag which specifies if the formatting has to be soft
  or hard, the number of columns of the desired text and possibly
  what characters are to be considered indentation characters. If
  a number smaller than 1 is passed, the default size will be used
  (70). The indentation_characters can be NULL or a string with the
  characters which will be discarded during indentation detection
  in search of the "real" text.

  This function is used to create two threads, consumer and producer
  and initialise a common communication list with a mutex and
  condition variable.  The producer thread will be reading data
  from STDIN, formatting input and waiting for a specific block of
  lines. When those lines are ready, it will fork a fmt process with
  two pipes. The output pipe will be communicated to the consumer
  thread via the shared list.

  Finalisation is communicated through the list as an object whose
  indentation level is negative or lines pointer NULL.
  
  Return: 0 with success, non-zero if there were problems.
  
***************************************************************************/

int email_fmt(int soft_format, int target_block_width,
    const char *indentation_characters)
{
    pthread_t producer, consumer;

    /* Store global session params. */
    if (target_block_width > 0)
        _target_block_width = target_block_width;

    if (indentation_characters)
        _indentation_characters = indentation_characters;

    if (init_shared_list()) {
        fprintf(stderr, "Couldn't initialise shared list\n");
        return 1;
    }

    if (pthread_create(&producer, 0, _producer_thread, (void*)soft_format) ||
            pthread_create(&consumer, 0, _consumer_thread, (void*)soft_format)) {
        fprintf(stderr, "Error creating threads\n");
        return 3;
    }

    if (pthread_join(producer, 0)) {
        fprintf(stderr, "Couldn't join on producer thread\n");
        return 4;
    }

    if (pthread_join(consumer, 0)) {
        fprintf(stderr, "Couldn't join on consumer thread\n");
        return 5;
    }

    /* Cleanup before exit. */
    destroy_shared_list();
    
    return 0;
    soft_format = 0;
}

/**[txh]********************************************************************

  Description: The producer thread reads from stdin looking for blocks of
  text with the same amount of indentation or EOF. When such a block is
  found, the indentation is removed and piped to a forked process. The
  stdout pipe of the forked process is stored in the global shared list
  for the consumer to process later.

  When all stdin has been consumed, the producer thread creates a dummy
  element in the shared list and exists. This is a signal to the consumer
  thread that there won't be any more input.

  The data parameter has to be a boolean softness parameter.
  
  Return: 0 on success.
  
***************************************************************************/

void *_producer_thread(void *data)
{
    READ_LINE buf;
    char *line, **lines = 0;
    const int soft_indentation = (int)data;
    int num_lines = 0;
    int indentation = -1;

    /* Start reading input. */
    init_read_line_buffer(&buf, STDIN_FILENO);
    
    /* Read lines until we reach EOF. */
    while ((line = read_line(&buf))) {
        /* Detect the indentation of the line. */
        if (indentation == -1) {
            /* First line read for the block. Just assign. */
            indentation = _count_indentation(line, soft_indentation);
        }
        else {
            const int new_level = _count_indentation(line, soft_indentation);
            /* Check if we need to "fork" the current block. */
            if (indentation != new_level) {
                /* Yes, we need to fork! */
                assert(indentation >= 0);
                _fork_line_block(lines, indentation);
                /* Ok, now clean the local array for another block. */
                lines = 0;
                num_lines = 0;
            }
            indentation = new_level;
        }
        
        /* Expand array of pointers and keep line. */
        lines = xrealloc(lines, sizeof(char *) * (num_lines + 2));
        lines[num_lines++] = line;
        lines[num_lines] = 0;
    }

    /* Do we have a pending array to format? */
    if (lines)
        _fork_line_block(lines, indentation);

    /* Ok, we finished for sure. Signal exit for consumer. */
    close(STDIN_FILENO);
    signal_dummy_object();
    
    return 0;
    data = 0;
}

/**[txh]********************************************************************

  Description: The consumer blocks on the global shared list waiting for
  incoming forked processes whose output has to be reindented before piping
  it out to stdout.

  The thread will end when a dummy element is found in the shared list. This
  dummy object contains a negative indentation level.
  
  The data parameter has to be a boolean softness parameter.
  
  Return: 0 on success.
  
***************************************************************************/

void *_consumer_thread(void *data)
{
    int indentation_level, process_stdout;
    const int soft_indentation = (int)data;
    char **lines;
    
    while (1) {
        lines = wait_for_input(&indentation_level, &process_stdout);

        if (!lines || indentation_level < 0) {
            /* Dummy object signalled. Close and exit. */
            close(STDOUT_FILENO);
            pthread_exit((void*)0);
        }
        else {
            int padding_length, f;
            char *line, padding[_target_block_width];
            READ_LINE buf;
            
            assert(lines);
            assert(lines[0]);
            assert(indentation_level < _target_block_width - 2);

            /* Generate correct padding. */
            if (indentation_level) {
                strncpy(padding, lines[0], indentation_level);
                if (soft_indentation) {
                    /* Finish string with null terminator. */
                    padding[indentation_level] = 0;
                }
                else {
                    /* With hard indentation, append a space. */
                    padding[indentation_level] = ' ';
                    padding[indentation_level + 1] = 0;
                }
            }
            else
                padding[0] = 0;

            padding_length = strlen(padding);
                
            init_read_line_buffer(&buf, process_stdout);
            /* Read all lines, piping them directly to the output. */
            while ((line = read_line(&buf))) {
                write(STDOUT_FILENO, padding, padding_length);
                write(STDOUT_FILENO, line, strlen(line));
            }

            /* Close child stdout file descriptor. */
            close(process_stdout);

            /* Free original lines' memory passed by producer thread. */
            for (f = 0; lines[f]; f++)
                free(lines[f]);
            free(lines);
        }
    }
    data = 0;
}

/**[txh]********************************************************************

  Description: Given a line of text, reads how many allowed indentation
  characters are present before any other character or EOL. The soft
  parameter tells the function if the exact raw count of indentation
  characters is to be returned (true), or the number of indentation
  characters without trailing whitespace (false).
  
  Return: Number of indentation characters used.
  
***************************************************************************/

static int _count_indentation(const char *line, const int soft)
{
    if (soft)
        return strspn(line, _indentation_characters);
    else {
        int num = strspn(line, _indentation_characters);
        /* We still need to "erase" those which are trailing whitespace. */

        while (num > 0 && isspace(line[num - 1]))
            num--;
            
        return num;
    }
}

/**[txh]********************************************************************

  Description: Here's the funny fork'ing thing. In lines is stored an array
  of pointers to lines of text which have the specified indentation. So
  the function creates two pipes, the lines and indentation level in the
  global list of communication with the consumer thread, then forks the
  current thread to run as a fmt command.
  
***************************************************************************/

static void _fork_line_block(char **lines, const int indentation)
{
    int max_length, child, to_child[2], from_child[2];
    assert(lines);
    assert(lines[0]);
    assert(indentation >= 0);

    /* Calculate the length we will pass to the fmt command. */
    if (_target_block_width - indentation < 20)
        max_length = 20;
    else
        max_length = _target_block_width - indentation;
        
    /* First create the communication pipes. Both half-duplex. */
    if (pipe(to_child) || pipe(from_child)) {
        fprintf(stderr, "Error creating pipes for forked process\n");
        pthread_exit((void*)5);
    }

    child = fork();
    
    if (child == 0) {
        /* We are the child process. */

        close(to_child[1]);
        dup2(to_child[0], 0); /* Standard input from one pipe. */
        close(from_child[0]);
        dup2(from_child[1], 1); /* Standard output to the other pipe. */
        /* Now the child has input and output redirected to the father. */

        /* Prepare the parameter and run external program. */
        char param_buf[15];
        snprintf(param_buf, sizeof(param_buf), "%d", max_length);
        /* fprintf(stderr, "Calling with %s\n", param_buf); */
        execlp("fmt", "fmt", "-w", param_buf, "-u", "-", NULL);
    }
    else if (child < 0) {
        /* There was an error. Abort. */
        fprintf(stderr, "Error creating child process\n");
        pthread_exit((void*)child);
    }
    else {
        int f;
        
        /* We are the parent. Prepare pipe and leave note to consumer. */
        close(to_child[0]);
        close(from_child[1]);
        queue_process(indentation, from_child[0], lines);
        
        /* We write to to_child[1] the rest of the lines. But how
           do we actually write them? The idea is quite easy, if the
           line contains ONLY whitespace characters, it is written
           as is. However, if there are other characters, the line
           is written using that first non whitespace character as
           the beginning. */
           
        for (f = 0; lines[f]; f++) {
            char *p = lines[f] + indentation;
            const int whitespace_len = strspn(p, _whitespace);
            const int rest_len = strlen(p + whitespace_len);

            if (rest_len > 0)
                write(to_child[1], p + whitespace_len, rest_len);
            else
                write(to_child[1], p, whitespace_len);
        }
            
        /* We are finished, close write file descriptor. */
        close(to_child[1]);
    }
}

