// -*- mode:C++; tab-width: 4 -*-
/*
    Multi-threaded C version fo email-fmt.py.
    Copyright (C) 2005 Grzegorz Adam Hankiewicz.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License,
    or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
    USA.
*/
#include "misc.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/**[txh]********************************************************************

  Description: Initialises a READ_LINE structure and prepares it
  for line reading.

  Example:

    READ_LINE buf;
    ...
    init_read_line_buffer(&buf, file_descriptor);
  
***************************************************************************/

void init_read_line_buffer(READ_LINE *buf, int file_descriptor)
{
    assert(buf);

    buf->fileno = file_descriptor;
    buf->start = buf->end;
    buf->buf[READ_LINE_STATIC_BUFFER] = 0;
}

/**[txh]********************************************************************

  Description: Given an initialised READ_LINE structure, reads
  lines from the associated file descriptor until EOF is reached
  or an error happens.
  
  Return: A pointer to the line (you have to free it). NULL if you
  reached the end of the file or an error happened.
  
***************************************************************************/

char *read_line(READ_LINE *buf)
{
    char *line = 0;
    int line_size = 0;
    assert(buf);

    while (1) {
        char *end;
        
        /* Read a new chunk from the input file if needed. */
        if (buf->start == buf->end) {
            int ret = read(buf->fileno, buf->buf, READ_LINE_STATIC_BUFFER);

            /* Did we reach EOF or had an input error? */
            if (ret < 1)
                return line;

            /* Reset static buffer pointers. */
            buf->start = 0;
            buf->end = ret;
            buf->buf[ret] = 0;
        }

        /* Find end of line. */
        end = strchr(buf->buf + buf->start, '\n');

        /* Did we find an end of line before the end of the static buffer? */
        if (!end) {
            /* No, we have to reallocate the line and continue reading. */
            int add_length = 1 + (buf->end - buf->start);
            line = xrealloc(line, line_size + add_length);
            
            /* Copy the new block, including null terminator, just in case. */
            strncpy(line + line_size, buf->buf + buf->start, add_length);
            line_size += add_length - 1;

            /* Reset the static pointers so next time we read more data. */
            buf->start = buf->end;
        }
        else {
            /* Yes, end of line was found. Copy it and return. */
            int add_length = end - (buf->buf + buf->start) + 1;
            line = xrealloc(line, line_size + add_length + 1);

            /* Copy the new block, and put an artificial null terminator. */
            strncpy(line + line_size, buf->buf + buf->start, add_length);
            line_size += add_length;
            line[line_size] = 0;

            buf->start += add_length;
            assert(buf->start <= buf->end);
            return line;
        }
    }
}

/**[txh]********************************************************************

  Description: Little helper to avoid checking errors of realloc.
  
  Return: Like stdlib realloc.
  
***************************************************************************/

void *xrealloc(void *ptr, size_t size)
{
    register void *value = realloc (ptr, size);
    if (value == 0) {
        fprintf(stderr, "Virtual memory exhausted");
        abort();
    }
    return value;
}
