// -*- mode:C++; tab-width: 4 -*-
/*
    Multi-threaded C version fo email-fmt.py.
    Copyright (C) 2005 Grzegorz Adam Hankiewicz.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License,
    or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
    USA.
*/
#include "shared-list.h"

#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>


/* Double linked list data. */
typedef struct shared_list_t SHARED_LIST;

struct shared_list_t {
    SHARED_LIST *prev, *next;
    /* Amount of reidentation to be performed. */
    int indentation_level;
    /* Pipe descriptor for the forked process. */
    int stdout;
    /* Original source lines. */
    char **lines;
};

/* Internal variables for synchronisation. */
static pthread_mutex_t _list_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t _pending_data = PTHREAD_COND_INITIALIZER;
static SHARED_LIST *_shared_list = 0;

static void _append_new_unit(SHARED_LIST *element);
static void _atfork_prepare(void);
static void _atfork_parent(void);
static void _atfork_child(void);

/**[txh]********************************************************************

  Description: Initialises the shared list, creating a dummy node
  that will hold both the start and end of the list. Also, registers atfork
  handlers to avoid having threading problems during fork calls.
  
  Return: 0 on success, non-zero otherwise.
  
***************************************************************************/

int init_shared_list(void)
{
    assert(!_shared_list && "Can't init twice a global resource!");
    
    /* Allocate memory. */
    _shared_list = malloc(sizeof(SHARED_LIST));
    if (!_shared_list)
        return 1;

    /* Prepare pointers. */
    _shared_list->next = _shared_list->prev = _shared_list;

    if (pthread_atfork(_atfork_prepare, _atfork_parent, _atfork_child))
        return 2;
        
    return 0;
}

/**[txh]********************************************************************

  Description: Cleans up the memory used by the shared list. Since this is
  going to be called by the consumer thread, the list is expected to be
  empty.
  
***************************************************************************/

void destroy_shared_list(void)
{
    assert(_shared_list);
    assert(_shared_list->next == _shared_list->prev);
    assert(_shared_list->next == _shared_list);
    
    free(_shared_list);
    _shared_list = 0;
}

/**[txh]********************************************************************

  Description: Used by the producer when no more inputs are going
  to be put on the shared list. This actually puts a dummy object
  on the list for the consumer to receive.
  
***************************************************************************/

void signal_dummy_object(void)
{
    SHARED_LIST *temp;
    assert(_shared_list);

    /* Create dummy object. */
    temp = malloc(sizeof(SHARED_LIST));
    if (!temp) {
        fprintf(stderr, "Not enough memory for dummy object!\n");
        pthread_exit((void*)1);
    }
    temp->indentation_level = -1;
    temp->stdout = -1;
    temp->lines = 0;

    _append_new_unit(temp);
}

/**[txh]********************************************************************

  Description: Waits for any input in the shared list. When this
  input is ready, it will be removed, assigned to the parameters
  passed and freed.
  
  Return: The original source lines left by the producer thread,
  plus both parameters are considered "output" parameters.
  
***************************************************************************/

char **wait_for_input(int *indentation_level, int *process_stdout)
{
    char **lines;
    SHARED_LIST *temp;
    assert(indentation_level && process_stdout);
    assert(_shared_list);
    
    pthread_mutex_lock(&_list_mutex);

    /* Wait until there is input. */
    while (_shared_list->next == _shared_list)
        pthread_cond_wait(&_pending_data, &_list_mutex);

    /* Now remove the input as fast as possible to release lock. */
    temp = _shared_list->next;
    temp->next->prev = _shared_list;
    _shared_list->next = temp->next;

    if (pthread_mutex_unlock(&_list_mutex)) {
        fprintf(stderr, "Error unlocking mutex during wait for input\n");
        pthread_exit((void*)3);
    }

    /* Ok, now return the values and free the memory. */
    *indentation_level = temp->indentation_level;
    *process_stdout = temp->stdout;
    lines = temp->lines;
    free(temp);
    return lines;
}

/**[txh]********************************************************************

  Description: Function used by the producer to queue data for the consumer
  thread. Pass the indentation level, the pipe to read from the forked
  process and the source lines of text.
  
***************************************************************************/

void queue_process(int indentation_level, int process_stdout, char **lines)
{
    SHARED_LIST *temp;
    assert(indentation_level >= 0);
    assert(lines);
    assert(lines[0]);
    assert(process_stdout >= 0);
    assert(_shared_list);

    /* Create new block for the queue. */
    temp = malloc(sizeof(SHARED_LIST));
    if (!temp) {
        fprintf(stderr, "Not enough memory for queueing process\n");
        pthread_exit((void*)7);
    }

    temp->lines = lines;
    temp->stdout = process_stdout;
    temp->indentation_level = indentation_level;

    /* Ok, now lock the list and insert at the end. */
    _append_new_unit(temp);
}

/**[txh]********************************************************************

  Description: Reusable bit of code which appends an element to the list
  of shared resources respecting mutex locks and signaling locked threads.
  
***************************************************************************/

static void _append_new_unit(SHARED_LIST *element)
{
    assert(element);
    
    pthread_mutex_lock(&_list_mutex);

    /* Relink as last element of the shared list. */
    element->prev = _shared_list->prev;
    element->next = _shared_list;
    _shared_list->prev->next = element;
    _shared_list->prev = element;

    /* Tell the consumer there is data ready. */
    pthread_cond_signal(&_pending_data);
    
    if (pthread_mutex_unlock(&_list_mutex)) {
        fprintf(stderr, "Error unlocking mutex after appending\n");
        pthread_exit((void*)2);
    }
}

/**[txh]********************************************************************

  Description: Locks all the mutexes to avoid threads modifying shared
  variables. Called by the parent just before a fork.
  
***************************************************************************/

static void _atfork_prepare(void)
{
    pthread_mutex_lock(&_list_mutex);
}

/**[txh]********************************************************************

  Description: Unlocks all the mutexes. Called by the parent just after a
  fork.
  
***************************************************************************/

static void _atfork_parent(void)
{
    if (pthread_mutex_unlock(&_list_mutex)) {
        fprintf(stderr, "Error unlocking mutex after fork()\n");
        abort();
    }
}

/**[txh]********************************************************************

  Description: Resets all the shared variables in the child process just
  after a fork.
  
***************************************************************************/

static void _atfork_child(void)
{
    pthread_mutex_init(&_list_mutex, 0);
    pthread_cond_init(&_pending_data, 0);
}

